from contextlib import contextmanager, suppress
from contextvars import ContextVar
from typing import Generator, Protocol

from fastapi import WebSocket

from jararaca.presentation.websocket.base_types import WebSocketMessageBase


class WebSocketConnectionManager(Protocol):

    async def broadcast(self, message: bytes) -> None: ...
    async def send(self, rooms: list[str], message: WebSocketMessageBase) -> None: ...
    async def join(self, rooms: list[str], websocket: WebSocket) -> None: ...
    async def add_websocket(self, websocket: WebSocket) -> None: ...
    async def remove_websocket(self, websocket: WebSocket) -> None: ...


_ws_manage_ctx = ContextVar[WebSocketConnectionManager]("ws_manage_ctx")


def use_ws_manager() -> WebSocketConnectionManager:
    try:
        return _ws_manage_ctx.get()
    except LookupError:
        raise RuntimeError("No WebSocketConnectionManager found")


@contextmanager
def provide_ws_manager(
    ws_manager: WebSocketConnectionManager,
) -> Generator[None, None, None]:
    token = _ws_manage_ctx.set(ws_manager)
    try:
        yield
    finally:
        with suppress(ValueError):
            _ws_manage_ctx.reset(token)



from contextlib import contextmanager, suppress
from contextvars import ContextVar
from typing import Any, Awaitable, Callable, TypeVar, Union, cast

DECORATED = TypeVar("DECORATED", bound=Union[Callable[..., Awaitable[Any]], type])


metadata_context: ContextVar[dict[str, Any]] = ContextVar("metadata_context")


def get_metadata(key: str) -> Any | None:
    return metadata_context.get({}).get(key)


@contextmanager
def provide_metadata(metadata: dict[str, Any]) -> Any:

    current_metadata = metadata_context.get({})

    token = metadata_context.set({**current_metadata, **metadata})
    try:
        yield
    finally:
        with suppress(ValueError):
            metadata_context.reset(token)


class SetMetadata:
    def __init__(self, key: str, value: Any) -> None:
        self.key = key
        self.value = value

    METATADA_LIST = "__metadata_list__"

    @staticmethod
    def register_metadata(cls: DECORATED, value: "SetMetadata") -> None:
        metadata_list = getattr(cls, SetMetadata.METATADA_LIST, [])
        metadata_list.append(value)
        setattr(cls, SetMetadata.METATADA_LIST, metadata_list)

    @staticmethod
    def get(cls: type) -> "list[SetMetadata]":
        return cast(list[SetMetadata], getattr(cls, SetMetadata.METATADA_LIST, []))

    def __call__(self, cls: DECORATED) -> DECORATED:
        SetMetadata.register_metadata(cls, self)
        return cls



import inspect
from typing import Any, Callable, TypeVar, cast

DECORATED_FUNC = TypeVar("DECORATED_FUNC", bound=Callable[..., Any])


class ScheduledAction:
    SCHEDULED_ACTION_ATTR = "__scheduled_action__"

    def __init__(
        self,
        cron: str,
        allow_overlap: bool = False,
        exclusive: bool = False,
        timeout: int | None = None,
        exception_handler: Callable[[BaseException], None] | None = None,
    ) -> None:
        """
        :param cron: A string representing the cron expression for the scheduled action.
        :param allow_overlap: A boolean indicating if the scheduled action should new executions even if the previous one is still running.
        :param exclusive: A boolean indicating if the scheduled action should be executed in one instance of the application. (Requires a distributed lock provided by a backend)
        """
        self.cron = cron
        self.allow_overlap = allow_overlap
        self.exclusive = exclusive
        self.exception_handler = exception_handler
        self.timeout = timeout

    def __call__(self, func: DECORATED_FUNC) -> DECORATED_FUNC:
        ScheduledAction.register(func, self)
        return func

    @staticmethod
    def register(func: DECORATED_FUNC, scheduled_action: "ScheduledAction") -> None:
        setattr(func, ScheduledAction.SCHEDULED_ACTION_ATTR, scheduled_action)

    @staticmethod
    def get_scheduled_action(func: DECORATED_FUNC) -> "ScheduledAction | None":
        if not hasattr(func, ScheduledAction.SCHEDULED_ACTION_ATTR):
            return None

        return cast(
            ScheduledAction, getattr(func, ScheduledAction.SCHEDULED_ACTION_ATTR)
        )

    @staticmethod
    def get_type_scheduled_actions(
        instance: Any,
    ) -> list[tuple[Callable[..., Any], "ScheduledAction"]]:

        members = inspect.getmembers(instance, predicate=inspect.ismethod)

        scheduled_actions: list[tuple[Callable[..., Any], "ScheduledAction"]] = []

        for _, member in members:
            scheduled_action = ScheduledAction.get_scheduled_action(member)

            if scheduled_action is None:
                continue

            scheduled_actions.append((member, scheduled_action))

        return scheduled_actions
